using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SimpleJSON;
using System.Linq;

namespace ClockwiseSilver
{
	public class SilverCircler : MVRScript
	{
		private JSONStorableBool isActive;
		private JSONStorableBool yzFlip;
		private JSONStorableFloat radiusJSON;
		private JSONStorableFloat speedXJSON;
		private JSONStorableFloat speedYJSON;
		private JSONStorableStringChooser refAtomJSON;
		private JSONStorableStringChooser targetJSON;
		private JSONStorableStringChooser refTargetJSON;
		private UIDynamicSlider ySlider;
		
		private FreeControllerV3 controller;
		private Rigidbody controlRB;
		private Atom refAtom;
		private Transform controllerTransform;
		private Transform referenceRBTransform;
		private Rigidbody referenceRB;
		private Vector3 controllerStartPos = Vector3.zero;
		private Vector3 currentRot = Vector3.zero;
		private Vector3 refPos = Vector3.zero;
		private Vector3 refDir = Vector3.forward;
		private float speedX = 0f;
		private float speedY = 0f;
		private bool animateEnabled = false;
		private bool useRef = false;
		
		
		public override void Init()
		{
			try
			{
				if (containingAtom.type != "Person")
				{
					SuperController.LogMessage("SilverCircler: Only works when added to a Person atom.");
					var h = CreateTextField(new JSONStorableString("header0", "Only works when added to a Person atom."),false);
					this.enabled = false;
					return;
				}
				
				isActive = new JSONStorableBool("isActive", false);
				isActive.storeType = JSONStorableParam.StoreType.Full;
				var toggle = CreateToggle(isActive, false);
				toggle.label = "Active";
				RegisterBool(isActive);

				targetJSON = new JSONStorableStringChooser("targetRB", null, null, "Circler", SyncTarget);
				RegisterStringChooser(targetJSON);
				List<string> choices = new List<string>();
				choices.Add("lHand");
				choices.Add("rHand");
				choices.Add("lFoot");
				choices.Add("rFoot");
				choices.Add("lThigh");
				choices.Add("rThigh");
				choices.Add("hip");
				choices.Add("pelvis");
				choices.Add("chest");
				choices.Add("neck");
				choices.Add("head");
				choices.Add("abdomen");
				choices.Add("abdomen2");
				targetJSON.choices = choices;
				var fp = CreateFilterablePopup(targetJSON);
				fp.popupPanelHeight = 700f;
				
				var space = CreateSpacer();
				space.height = 20f;
				var heading = CreateTextField(new JSONStorableString("header1",
					"Reference Atom: Optional\nMaintains the Circler's position\nrelative to another body."),false);
                heading.height = 40f;	
				
				//RefAtom
				refAtomJSON = new JSONStorableStringChooser("refAtom", SuperController.singleton.GetAtomUIDs(), null, "Ref Atom", RefTargetChoices);
				RegisterStringChooser(refAtomJSON);
				AtomChocies();
				fp = CreateFilterablePopup(refAtomJSON);
				fp.popupPanelHeight = 700f;
				fp.popup.onOpenPopupHandlers += AtomChocies;

				//RefTarget
				refTargetJSON = new JSONStorableStringChooser("refTarget", null, null, "Ref Target", SyncRefTarget);
				RegisterStringChooser(refTargetJSON);
				fp = CreateFilterablePopup(refTargetJSON);
				fp.popupPanelHeight = 600f;
				
				radiusJSON = NewSlider("Radius", 50f, -5000f, 5000f, true);
				speedXJSON = NewSlider("Speed X", 0f, -1000f, 1000f, true);
				speedYJSON = new JSONStorableFloat("Speed Y", 0f, -1000f, 1000f, true, true);
				speedYJSON.storeType = JSONStorableParam.StoreType.Full;
				ySlider = CreateSlider(speedYJSON, true);
				ySlider.label = "Speed Y";
				RegisterFloat(speedYJSON);
				
				space = CreateSpacer(true);
				space.height = 20f;
				heading = CreateTextField(new JSONStorableString("header2",
					"Reference Direction (YZ):\nChanges the second rotation axis\nfrom Y to Z."),true);
                heading.height = 40f;

				//RefDir
				yzFlip = new JSONStorableBool("yzFlip", false);
				yzFlip.storeType = JSONStorableParam.StoreType.Full;
				toggle = CreateToggle(yzFlip, true);
				toggle.label = "Y Z Flip";
				RegisterBool(yzFlip);		
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}
		
		private void Start()
		{
			isActive.toggle.onValueChanged.AddListener(checkVal =>
				{
					if (checkVal) { StartCircler(); }
					else { StopCircler(); }
				});
				
			yzFlip.toggle.onValueChanged.AddListener(checkVal =>
				{
					SyncRefDir(checkVal);
				});
			SyncRefDir(yzFlip.val);
			
			speedXJSON.slider.onValueChanged.AddListener((value) => { SyncSpeed(); });
			speedYJSON.slider.onValueChanged.AddListener((value) => { SyncSpeed(); });
			SyncSpeed();
			
			if (isActive.val) { StartCircler(); }
		}
		
		private void StartCircler()
		{
			if (controller == null || controlRB == null)
			{
				SuperController.LogMessage("SilverCircler: The circler is not a valid target.");
				animateEnabled = false;
				isActive.val = false;
				return;
			}
			else
			{
				currentRot = Vector3.zero;
				controllerStartPos = controllerTransform.position;
				if (referenceRB != null && refTargetJSON.val != "None")
				{
					useRef = true;
					refPos = referenceRBTransform.InverseTransformPoint(controllerTransform.position);
				}
				else
				{
					useRef = false;
					refPos = controllerTransform.position;
				}
				animateEnabled = true;
			}
		}
		
		private void StopCircler()
		{
			if (!animateEnabled) { return; }
			if (!useRef)
			{
				controllerTransform.position = controllerStartPos;
			}
			animateEnabled = false;
		}
		
		private float Rollover(float x)
		{
			if (x > 360f) { return x - 360f; }
			else if (x < 0f) { return 360f - x; }
			return x;
		}
		
		private void FixedUpdate()
		{
			try
			{
				if (animateEnabled)
				{
					float dt = Time.fixedDeltaTime;
					currentRot.x = Rollover(currentRot.x + (speedX * dt));
					if (yzFlip.val)
					{
						currentRot.z = Rollover(currentRot.z + (speedY * dt));
					}
					else
					{
						currentRot.y = Rollover(currentRot.y + (speedY * dt));
					}
					
					Vector3 dir = controllerTransform.TransformDirection(Quaternion.Euler(currentRot) * refDir);
					if (useRef)
					{
						controllerTransform.position = referenceRBTransform.TransformPoint(refPos);
					}
					
					controlRB.AddForce(dir * radiusJSON.val, ForceMode.Force);
				}
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}			
		}
		
		private void SyncTarget(string targetID)
		{
			if (animateEnabled) { isActive.val = false; }
			if (targetID != null && targetID != "" && targetID != "None")
			{
				controlRB = containingAtom.rigidbodies.First(rb => rb.name == targetID);
				controller = containingAtom.freeControllers.First(fc => fc.name == targetID + "Control");
				controllerTransform = controller.transform;
			}
			else
			{
				controlRB = null;
				controller = null;
			}
		}
		
		private void AtomChocies()
		{
			if (animateEnabled) { isActive.val = false; }
			List<string> atomChoices = new List<string>();
			atomChoices.Add("None");
			foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
			{
				atomChoices.Add(atomUID);
			}
			refAtomJSON.choices = atomChoices;
		}
		
		private void RefTargetChoices(string atomUID)
		{
			if (animateEnabled) { isActive.val = false; }
			List<string> choices = new List<string>();
			choices.Add("None");
			if (atomUID != null)
			{
				refAtom = SuperController.singleton.GetAtomByUid(atomUID);
				if (refAtom != null)
				{
					foreach (Rigidbody rb in refAtom.rigidbodies)
					{
						if (!rb.name.Contains("Collider") && !rb.name.Contains("Control") && !rb.name.Contains("Link"))
						{
							choices.Add(rb.name);
						}
					}
				}
			}
			else
			{
				refAtom = null;
			}
			refTargetJSON.choices = choices;
		}
		
		private void SyncRefTarget(string refTargetID)
		{
			if (animateEnabled) { isActive.val = false; }
			if (refTargetID != null && refTargetID != "" && refTargetID != "None")
			{
				referenceRB = refAtom.rigidbodies.First(rb => rb.name == refTargetID);
				referenceRBTransform = referenceRB.transform;
			}
			else
			{
				referenceRB = null;
			}
		}
		
		private void SyncRefDir(bool checkVal)
		{
			if (animateEnabled) { isActive.val = false; }
			if (checkVal) { ySlider.label = "Speed Z"; refDir = Vector3.up; }
			else { ySlider.label = "Speed Y"; refDir = Vector3.forward; }
		}
		
		private void SyncSpeed()
		{
			speedX = speedXJSON.val;
			speedY = speedYJSON.val;
		}
		
		private JSONStorableFloat NewSlider(string name, float val, float min, float max, bool right)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, val, min, max, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full; CreateSlider(storable, right); RegisterFloat(storable);
			return storable;
		}
	}
}